/***
    Copyright (C) 2013 Cable Developers

    This program or library is free software; you can redistribute it
    and/or modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with this library; if not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301 USA.
***/

public class Cable.Controller.Channel : GLib.Object {

    public weak Server server;
    public View.Channel view;
    public Irc.Channel backend;

    public string name; //TODO id

    public Channel (Server server, string name) {
        this.server = server;
        this.name = name;

        if (server.view.has_chat (name) == false)
            server.view.add_channel (name, name);

        view = server.view.get_chat (name) as View.Channel;
        view.nick = server.nick;
        backend = new Irc.Channel (name, server.backend);

        connect_frontend_signals ();
        connect_backend_signals ();

        load_state ();
    }

    void load_state () {
        /*if (backend.joined) {
            view.room.state = Widgets.Room.State.CHAT;
            FIXME some bug in maki returns garbage here:
            -> view.room.list_users (backend.nicks, backend.prefixes);
            backend.event_topic (backend.topic, "");
        } else {
            view.room.state = Widgets.Room.State.JOINING;
        }*/
        // part and rejoin to workaround bug
        view.state = State.JOINING;

        if (server.backend.connected == false) {
            server.backend.event_connected.connect (() => {
                if (backend.joined)
                    backend.send_part ();
                backend.send_join ();
            });
        } else {
            if (backend.joined)
                backend.send_part ();
            backend.send_join ();
        }
    }

    // GUI signals
    void connect_frontend_signals () {
        view.part.connect (() => {
            backend.send_part ();
        });

        server.window.view.selected.connect ((item) => {
            if (item == view)
                view.unread_messages = 0;
        });

        view.join.connect (() => {
            view.state = State.JOINING;
            backend.send_join ();
        });

        view.send_message.connect ((message) => {
            if (message.has_prefix ("/"))
                send_command (message.replace ("/", ""));
            else
                server.backend.send_message (name, message);

            view.entry = "";
        });

        view.focused.connect (() => {
            Services.Launcher.jobs = 0;
            Services.Launcher.urgent = false;
            return false;
        });

        view.private_chat_requested.connect ((name) => {
            server.open_new_chat_with (name);
            if (server.private_chats.has_key (name))
                server.window.view.current_chat = server.private_chats [name].view;
        });
    }

    // IRC signals
    void connect_backend_signals () {
        backend.event_message.connect ((from, message) => {
            if (server.window.view.current_chat != view)
                view.unread_messages++;
        });

        backend.event_topic.connect ((from, topic) => {
            view.topic = topic;
        });

        server.backend.event_nick.connect ((from, new_nick) => {
            if (server.backend.is_me (from)) {
                view.nick = new_nick;
            }

            assert (new_nick != null);

            var name = from.split ("!")[0];

            if (!view.has_user (name))
                return;

            if (name.has_prefix ("@"))
                view.rename_user (name.replace ("@", ""), new_nick, UserType.OPERATOR);
            else if (name.has_prefix ("+"))
                view.rename_user (name.replace ("+", ""), new_nick, UserType.VOICED);
            else
                view.rename_user (name, new_nick, UserType.REGULAR);
        });

        backend.event_part.connect ((from, message) => {
            if (server.backend.is_me (from)) {
               view.state = State.NOT_JOINED;
            } else {
                var name = from.split ("!")[0];

                if (name.has_prefix ("@"))
                    view.remove_user (name.replace ("@", ""), message);
                else if (name.has_prefix ("+"))
                    view.remove_user (name.replace ("+", ""), message);
                else
                    view.remove_user (name, message);
            }
        });

        backend.event_join.connect ((from) => {
            if (server.backend.is_me (from)) {
                view.state = State.CHAT;
            } else {
                var name = from.split ("!")[0];

                if (name.has_prefix ("@"))
                    view.add_user (name.replace ("@", ""), UserType.OPERATOR);
                else if (name.has_prefix ("+"))
                    view.add_user (name.replace ("+", ""), UserType.VOICED);
                else
                    view.add_user (name, UserType.REGULAR);
            }
        });

        backend.event_no_such.connect (() => {
            view.state = State.NO_SUCH;
        });

        backend.event_message.connect ((from, message) => {
            var nick = from.split ("!") [0];

            if (message.contains (server.nick)) {

                if (view.has_focus == false) {
                    Services.Notify.send (nick, message);
                    Services.Launcher.jobs++;
                    Services.Launcher.urgent = true;
                }

                view.message (nick, message, MessageType.PING);
            } else {
                view.message (nick, message);
            }
        });

        backend.event_action.connect ((from, message) => {
            var nick = from.split ("!")[0];
            view.message ("", "* " + nick + " " + message, MessageType.ACTION);
        });

        backend.event_names.connect (view.list_users);
    }

    void send_command (string command) {

        var args = command.split (" ");

        switch (args[0]) {
            case "join":
                return_if_fail (args.length > 1);
                var channel_with_hash = args[1];
                if (channel_with_hash.slice(0, 1) != "#") {
                    channel_with_hash = "#" + args[1];
                }
                var channel = new Channel (server, channel_with_hash);
                server.channels[channel_with_hash] = channel;
                server.window.view.current_chat = channel.view;
                break;

            case "part":
                backend.send_part ();
                break;

            case "quit":
                server.backend.send_quit ();
                break;

            case "nick":
                return_if_fail (args.length > 1);
                server.backend.send_nick (args[1]);
                break;

            case "topic":
                return_if_fail (args.length > 1);
                backend.send_topic (string.joinv (" ", args[1:-1]));
                break;

            case "me":
            case "action":
                return_if_fail (args.length > 1);
                backend.send_action (string.joinv (" ", args[1:-1]));
                break;

            case "kick":
                return_if_fail (args.length > 2);
                backend.send_kick (args[1], string.joinv (" ", args[2:-1]));
                break;

            case "ignore":
                return_if_fail (args.length > 1);
                server.backend.send_ignore (string.joinv (" ", args[1:-1]));
                break;

            case "unignore":
                return_if_fail (args.length > 1);
                server.backend.send_unignore (string.joinv (" ", args[1:-1]));
                break;

            case "msg":
                return_if_fail (args.length > 2);
                server.backend.send_message (args[1], string.joinv (" ", args[2:-1]));
                break;

            case "notice":
                return_if_fail (args.length > 2);
                server.backend.send_notice (args[1], string.joinv (" ", args[2:-1]));
                break;

            case "away":
                if (args.length < 2)
                    server.backend.send_back ();
                else
                    server.backend.send_away (string.joinv (" ", args[1:-1]));
                break;

            default:
                server.backend.send_raw (command);
                break;
        }
    }
}
