/***
    Copyright (C) 2013 Cable Developers

    This program or library is free software; you can redistribute it
    and/or modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with this library; if not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301 USA.
***/

public class Cable.Controller.Window : GLib.Object {

    public View.Window view { get; set; }

    public Gee.HashMap <string, Server> servers {
        get; set;
        default = new Gee.HashMap <string, Server> ();
    }

    public Window (View.Window window) {
        GLib.Object (view: window);

        connect_frontend_signals ();
        connect_backend_signals ();
        initialize ();
    }

    // restore saved state
    void initialize () {
        var channels = Settings.preferences.channels;

        // check for joined channels
        foreach (var server_name in Irc.Client.known_server_names) {
            var server = new Irc.Server.from_name (server_name);
            foreach (var channel in server.channel_names)
                join_channel (server.address, server.nick, channel);
        }

        // check for channels from last session
        if (channels.length > 0) {
            foreach (var channel in channels) {
                var parts = channel.split (":");
                return_if_fail (parts.length == 3);
                join_channel (parts[0], parts[1], parts[2]);
            }
        }

        if (servers.size == 0)
            view.welcome_shown = true;
    }

    // backend events that affect the window's state
    void connect_backend_signals () {
        Irc.client.event_nick.connect ((time, server, from, nick) => {
            var current_chat = view.current_chat;
            var current_server = view.current_server;

            if (server == current_server.id)
                view.title = string.join (" — ", current_chat.name, nick);
        });
    }

    // frontend signals from the window
    void connect_frontend_signals () {
        view.selected.connect ((item) => {
            if (item is View.Channel) {
                var channel = item as View.Channel;
                view.title = string.join (" — ", channel.name);
                Settings.preferences.last_used_nick = channel.nick;
            }
        });

        view.join.connect ((a,n,c) => {
            join_channel (a,n,c);
        });

        view.remove_channel.connect (() => {
            var channel_id = view.current_chat.id;
            var server_id = view.current_server.id;

            if (view.current_chat == null) {
                warning ("No channel selected, cannot leave.");
            } else if (view.current_chat is View.Channel) {
                part_channel (servers[server_id].channels[channel_id]);
            } else if (view.current_chat is View.PrivateChat) {
                var channel = servers[server_id].private_chats[channel_id];
                var server = channel.server;
                server.view.remove_chat (channel.view.id);
                server.private_chats.unset (channel.view.id);

                if (server.channels.size == 0)
                    servers.unset (server.view.id);

                if (servers.size == 0) {
                    view.title = App.TITLE;
                    view.welcome_shown = true;
                }
            }
        });

        view.add_channel.connect (() => {
            var dialog = new Dialogs.JoinChannel ();
            dialog.done.connect ((a,b,c) => join_channel (a.address,b.nick_name,c));
            dialog.show_all ();
        });

        view.show_preferences.connect (() => {
            var dialog = new Dialogs.Preferences ();
            dialog.show_all ();
        });

        view.close.connect (() => {
            string[] chans = {};

            foreach (var server in servers.values) {
                foreach (var channel in server.channels.values) {
                    chans += string.join (":", server.address, server.nick, channel.name);
                    channel.backend.send_part ();
                }
            }

            Settings.preferences.channels = chans;
        });
    }

    void join_channel (string address, string nick, string channel_name) {
        Server server = null;
        foreach (var s in servers.values)
            if (s.address == address && s.nick == nick)
                server = s;

        if (server == null) {
            server = new Server (address, nick, this);
            servers[server.backend.name] = server;
        }

        if (server.view.has_chat (channel_name)) {
            warning ("Channel is already joined");
            return;
        }

        var channel = new Channel (server, channel_name);
        server.channels[channel_name] = channel;
        
        view.current_chat = channel.view;
        view.welcome_shown = false;
    }

    void part_channel (Channel channel) {
        var server = channel.server;

        channel.backend.send_part ();
        server.view.remove_chat (channel.view.id);
        server.channels.unset (channel.view.id);

        if (server.channels.size == 0)
            servers.unset (server.view.id);

        if (servers.size == 0) {
            view.title = App.TITLE;
            view.welcome_shown = true;
        }
    }
}
