/***
    Copyright (C) 2013 Cable Developers

    This program or library is free software; you can redistribute it
    and/or modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with this library; if not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301 USA.
***/

/**
 * Cable.Irc.ServerConfig is a class that allows you to change the configuration of a server.
 * This configuration is a normal keyfile. Possible keys are:
 *
 *  autoconnect (should always be true)
 *  port (default, maki takes care of this)
 *  ssl (-> network)
 *  nick (-> identity)
 *  user (automatic)
 *  name (-> identity)
 *  address (-> network)
 *  nickserv (-> server, I guess...)
 */
public class Cable.Irc.ServerConfig : GLib.Object {

    /**
     * The server's name. Used by maki.
     */
    public string name { get; set; }

    /**
     * Create a new ServerConfig object.
     */
    public ServerConfig (string name) {
        Object (name: name);
    }

    /**
     * Get the value of a given key.
     * @param key Key whose value you want to query.
     * @return Value of the key.
     */
    public new string @get (string key) {
        string result = "";

        try {
            result = client.server_get (this.name, "server", key);
        } catch (GLib.IOError error) {
            warning (error.message);
        }

        return result;
    }

    /**
     * Assign a new value to a given key.
     * @param key Key whose value you want to change.
     * @param value New value for the given key. Use null to unset the key.
     */
    public new void @set (string key, string? value) {
        if (value == null) {
            unset (key);
            return;
        }

        try {
            client.server_set (this.name, "server", key, value);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Remove a key from the config file.
     * @param key Key you want to remove.
     */
    public void unset (string key) {
        try {
            client.server_remove (this.name, "server", key);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }
}

/**
 * The Cable.Irc.Server class is an easy to use interface to the maki daemon and provides
 * methods and signals to communicate with maki via dbus.
 *
 * This class contains a static list of all servers. This ensures that there are no duplicates.
 * XXX
 */
public class Cable.Irc.Server : GLib.Object {

    public static Gee.ArrayList <Server> all { get; private set; }

    /**
     * Information about the network we are connected to.
     */
    public Utils.Network network {
        owned get {
            return Utils.Network.get_or_create (config["address"]);
        }
    }

    /**
     * Information about the identity we are using for this connection.
     */
    public Utils.Identity identity {
        owned get {
            return Utils.Identity.get_or_create (config["nick"]);
        }
    }
    
    public string address {
        owned get { return config["address"]; }
    }
    
    public string nick {
        owned get { return config["nick"]; }
    }

    /**
     * Server configuration.
     */
    public ServerConfig config { get; private set; }

    /**
     * The name maki uses to identify a server.
     */
    public string name { get; private set; }

    /**
     * Am I away?
     */
    public bool away {
        get {
            try {
                return client.user_away (this.name, identity.nick_name);
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return false;
        }
    }

    /**
     * Am I connected?
     */
    public bool connected {
        get {
            try {
                return this.name in client.servers ();
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return false;
        }
    }


    /**
     * List of joined channels.
     */
    public string[] channel_names {
        owned get {
            try {
                return client.channels (this.name);
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return {};
        }
    }

    /**
     * XXX
     */
    public string[] supported_prefixes {
         owned get {
            try {
                return client.support_prefix (this.name);
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return {};
        }
    }

    /**
     * XXX
     */
    public string supported_channel_types {
         owned get {
            try {
                return client.support_chantypes (this.name);
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return "";
        }
    }

    /**
     * XXX
     */
    public string[] ignores {
         owned get {
            try {
                return client.ignores (this.name);
            } catch (GLib.IOError error) {
                warning (error.message);
            }

            return {};
        }
    }

    /* TODO put in channel as property and use unlimited lines:
    public string[] log (string target, uint64 lines) throws GLib.IOError {
        return client.log (this.name, target, lines);
    }*/

    // TODO document this
    public signal void event_away ();
    public signal void event_away_message (string nick, string message);
    public signal void event_back ();
    public signal void event_connected ();
    public signal void event_error (string domain, string reason, string[] arguments);
    public signal void event_invite (string from, string channel, string who);
    public signal void event_list (string channel, uint64 users, string topic);
    public signal void event_message (string from, string target, string message);
    public signal void event_mode (string from, string target, string mode, string parameter);
    public signal void event_motd (string message);
    public signal void event_nick (string from, string new_nick);
    public signal void event_no_such (string target, string type);
    public signal void event_notice (string from, string target, string message);
    public signal void event_oper ();

    /**
     * Someone has quit a channel you are in.
     * @param from User who quit the channel.
     * @param message Message displayed when the user quits.
     */
    public signal void event_quit (string from, string message);

    /**
     * A users away status has been changed.
     * @param from Affected user.
     * @param away New status.
     */
    public signal void event_user_away (string from, bool away);
    public signal void event_whois (string nick, string message);

    /**
     * A server is uniquely identified by a combination of a network and an identity.
     */
    public Server (string address, string nick) {
    debug (@"+ IRC SERVER $address");
        setup_signals ();

        foreach (var server in Client.server_names) {
            var config = new ServerConfig (server);
            if (config["nick"] == nick &&
                config["address"] == address &&
                config["client"] == "cable") {
                name = server;
                break;
            }
        }

        if (name != null) {
            debug (@"Already connected to $name.");
            config = new ServerConfig (name);
        } else {
            name = "cable|" + GLib.Random.next_int ().to_string ();

            debug (@"Creating new server $name.");

            config = new ServerConfig (name);

            config["nick"] = nick;
            config["address"] = address;
            config["nickserv"] = network.password;
            config["name"] = identity.real_name;
            config["autoconnect"] = "true";
            config["client"] = "cable";

            debug ("Successfully set up server.\n");
        }

        all.add (this);

        event_nick.connect_after ((from, new_nick) => {
            if (is_me (from))
                config["nick"] = new_nick;
        });
    }

    /**
     * Creates a server from its maki name.
     */
    public Server.from_name (string name)
        requires (name in Client.known_server_names)
        requires (get_by_name (name) == null)
        ensures (get_by_name (name) == this) {
        
        setup_signals ();
        
        this.name = name;

        config = new ServerConfig (name);

    debug (@"+ IRC SERVER $address (from_name)");
    
        all.add (this);

        event_nick.connect_after ((from, new_nick) => {
            if (is_me (from))
                config["nick"] = new_nick;
        });
    }
    
    ~Server () { debug (@"- IRC SERVER $address"); }

    public static void init ()
        requires (all == null) {
        
        all = new Gee.ArrayList <Server> ();
    }

    /**
     * Get a server object by its name.
     * 
     */
    public static Server? get_by_name (string query) {
        foreach (var server in Server.all)
            if (server.name == query)
                return server;

        return (Server) null;
    }

    /**
     * Get the server that uses the given network and identity.
     * @param network The network the queried server is using.
     * @param identity The identity the queried server is using.
     * @return The queried server if it exists, otherwise null.
     */
    public static Server get_or_create (string address, string nick) {
        return search (address, nick) ?? new Server (address, nick);
    }
    
    public static Server? search (string address, string nick) {
        foreach (var server in Server.all)
            if (server.config["address"] == address &&
                server.config["nick"] == nick)
                return server;
        return (Server) null;
    }



    /**
     * Check if a command was issued by me.
     */
    public bool is_me (string from) {
        var nick = from.split ("!")[0];
        return nick == identity.nick_name;
    }

    /*void debug_message (string title, string args) {
        debug (@"[$(network.address) - $(identity.nick_name)] $title ($args)");
    }*/

    /**
     * Pass on signals from the backend daemon without the time and server parameters.
     */
    void setup_signals () {
        client.event_away.connect ((time, server) => {
            if (server == this.name)
                event_away ();
        });

        client.event_away_message.connect ((time, server, nick, message) => {
            if (server == this.name)
                event_away_message (nick, message);
        });

        client.event_back.connect ((time, server) => {
            if (server == this.name)
                event_back ();
        });

        client.event_connected.connect ((time, server) => {
            if (server == this.name)
                event_connected ();
        });

       /* client.event_ctcp.connect ((time, server, from, target, message) => {
            if (server == this.name)
                event_ctcp (from, target, message);
        });

        client.event_dcc_send.connect ((time, server, id, from, filename, size, progress, speed, status) => {
            if (server == this.name)
                event_dcc_send (id, from, filename, size, progress, speed, status);
        });*/

        client.event_error.connect ((time, server, domain, reason, arguments) => {
            if (server == this.name)
                event_error (domain, reason, arguments);
        });

        client.event_invite.connect ((time, server, from, channel, who) => {
            if (server == this.name)
                event_invite (from, channel, who);
        });

        client.event_list.connect ((time, server, channel, users, topic) => {
            if (server == this.name)
                event_list (channel, users, topic);
        });

        client.event_message.connect ((time, server, from, target, message) => {
            if (server == this.name)
                event_message (from, target, message);
        });

        client.event_mode.connect ((time, server, from, target, mode, parameter) => {
            if (server == this.name)
                event_mode (from, target, mode, parameter);
        });

        client.event_motd.connect ((time, server, message) => {
            if (server == this.name)
                event_motd (message);
        });

        client.event_nick.connect ((time, server, from, new_nick) => {
            if (server == this.name)
                event_nick (from, new_nick);
        });

        client.event_no_such.connect ((time, server, target, type) => {
            if (server == this.name)
                event_no_such (target, type);
        });

        client.event_notice.connect ((time, server, from, target, message) => {
            if (server == this.name)
                event_notice (from, target, message);
        });

        client.event_oper.connect ((time, server) => {
            if (server == this.name)
                event_oper ();
        });

        client.event_quit.connect ((time, server, from, message) => {
            if (server == this.name)
                event_quit (from, message);
        });

        client.event_user_away.connect ((time, server, from, away) => {
            if (server == this.name)
                event_user_away (from, away);
        });

        client.event_whois.connect ((time, server, nick, message) => {
            if (server == this.name)
                event_whois (nick, message);
        });
    }

    /**
     * Set your status to "away" and leave a message.
     * @param message Message displayed to users trying to talk to you.
     */
    public void send_away (string? message = null) {
        try {
            client.away (this.name, message ?? identity.away_message);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Unsets "away" status.
     */
    public void send_back () {
        try {
            client.back (this.name);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Connect to this server.
     */
    public new void send_connect () {
        try {
            client.connect (this.name);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /*public void ctcp (string target, string message) throws GLib.IOError {
        client.ctcp (this.name, target, message);
    }

    public void dcc_send (string target, string path) throws GLib.IOError {
        client.action (this.name, target, path);
    }

    public void dcc_sends (out uint64[] ids, out string[] servers, out string[] froms, out string[] filenames, out uint64[] sizes, out uint64[] progresses, out uint64[] speeds, out uint64[] statuses) throws GLib.IOError {
        client.action (this.name, channel, message);
    }

    public void dcc_send_accept (uint64 id) throws GLib.IOError {
        client.action (this.name, channel, message);
    }

    public string dcc_send_get (uint64 id, string key) throws GLib.IOError {
        client.action (this.name, channel, message);
    }

    public void dcc_send_remove (uint64 id) throws GLib.IOError {
        client.action (this.name, channel, message);
    }

    public void dcc_send_resume (uint64 id) throws GLib.IOError {
        client.action (this.name, channel, message);
    }

    public void dcc_send_set (uint64 id, string key, string value) throws GLib.IOError {
        client.action (this.name, channel, message);
    }*/

    /**
     * Ignore messages from users who fit the given pattern.
     * @param pattern The pattern to use.
     */
    public void send_ignore (string pattern) {
        try {
            client.ignore (this.name, pattern);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Send a message to a givent target.
     * @param target can be a channel or a user.
     * @param message Message you want to send.
     */
    public void send_message (string target, string message) {
        try {
            client.message (this.name, target, message);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    public void send_mode (string target, string mode) {
        try {
            client.mode (this.name, target, mode);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Send a NICK request to change your current nick on this network.
     * @param nick Your new nick name.
     */
    public void send_nick (string nick) {
        try {
            client.nick (this.name, nick);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Identify with nickserv.
     */
    public void send_nickserv () {
        try {
            client.nickserv (this.name);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Send a NOTICE.
     * @param target Can be a TODO ??
     * @param message The message you want to send.
     */
    public void send_notice (string target, string message) {
        try {
            client.notice (this.name, target, message);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Send OPER request.
     * @param name TODO
     * @param password OP password.
     */
    public void send_oper (string name, string password) {
        try {
            client.oper (this.name, name, password);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }
    
    public void send_part (string channel, string? message = null) {
        try {
            client.part (this.name, channel, message ?? identity.part_message);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Disconnect from the server and all channels.
     * @param message Message displayed to other users.
     */
    public void send_quit (string message = identity.quit_message) {
        try {
            client.quit (this.name, message);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Send a raw IRC command.
     * @param command The raw command you want to send.
     */
    public void send_raw (string command) {
        try {
            client.raw (this.name, command);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * Remove a certain pattern from the ignored list.
     * @param pattern The pattern.
     */
    public void send_unignore (string pattern) {
        try {
            client.unignore (this.name, pattern);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * TODO
     */
    public void send_who (string mask, bool operators_only) {
        try {
            client.who (this.name, mask, operators_only);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /**
     * TODO
     */
    public void send_whois (string mask) {
        try {
            client.whois (this.name, mask);
        } catch (GLib.IOError error) {
            warning (error.message);
        }
    }

    /*public string server_get (string group, string key) throws GLib.IOError {
        return client.server_get (this.name, group, key);
    }

    public string[] server_get_list (string group, string key) throws GLib.IOError {
        return client.server_get_list (this.name, group, key);
    }

    public string[] server_list (string group) throws GLib.IOError {
        return client.server_list (this.name, group);
    }

    public void server_remove (string group, string key) throws GLib.IOError {
        client.server_remove (this.name, group, key);
    }

    public void server_rename (string new_name) throws GLib.IOError {
        client.server_rename (this.name, new_name);
    }

    public void server_set (string group, string key, string value) throws GLib.IOError {
        client.server_set (this.name, group, key, value);
    }

    public void server_set_list (string group, string key, string[] list) throws GLib.IOError {
        client.server_set_list (this.name, group, key, list);
    }*/
}
