/***
    Copyright (C) 2013 Cable Developers

    This program or library is free software; you can redistribute it
    and/or modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with this library; if not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301 USA.
***/

public enum Cable.UserType {
    OPERATOR,
    VOICED,
    REGULAR
}

public enum Cable.State {
    CHAT,
    JOINING,
    NOT_JOINED,
    NO_SUCH;
}

/**
 * This class represents a chat room, i.e, the topic, the message display,
 * a list of users and is associated to a channel.
 * This class is always accessed through its related channel.
 */
public class Cable.Widgets.Room : Gtk.Grid {

    public UserList user_list;
    public ChatDisplay chat;
    public Gtk.Entry text_entry;
    public Widgets.Topic topic_bar;
    public Gtk.Label pseudo_label;
    
    Granite.Widgets.EmbeddedAlert not_joined_alert;
    Granite.Widgets.EmbeddedAlert no_such_alert;
    Granite.Widgets.EmbeddedAlert loading_screen;
    Granite.Widgets.ThinPaned paned;

    Utils.AutoScrolled left_scrolled;

    Gtk.Action action_join;

    public signal void send_message (string message);
    // public signal void request_complete ();
    public signal void rejoin ();

    public string channel { get; set; }

    public string nick {
        get { return pseudo_label.label; }
        set { pseudo_label.label = value; }
    }

    public string entry {
        get { return text_entry.text; }
        set { text_entry.text = value; }
    }

    public Utils.MessageHistory history { get; private set; }

    State _state;
    public State state {
        get { return _state; }
        set {
            _state = value;

            var widget = get_child_at (0, 0);
            if (widget != null)
                remove (widget);

            switch (_state) {
                case State.CHAT:
                    attach (paned, 0, 0, 1, 1);
                    break;
                case State.JOINING:
                    attach (loading_screen, 0, 0, 1, 1);
                    loading_screen.working = true;
                    break;
                case State.NOT_JOINED:
                    attach (not_joined_alert, 0, 0, 1, 1);
                    break;
                case State.NO_SUCH:
                    attach (no_such_alert, 0, 0, 1, 1);
                    break;
                default:
                    assert_not_reached ();
            }

            show_all ();
        }
    }

    public Room (string channel, string nick = "") {
        this.channel = channel;

        history = new Utils.MessageHistory ();

        init_layout ();
        connect_callbacks ();
    }

    void init_layout () {
        user_list = new UserList ();
        
        paned = new Granite.Widgets.ThinPaned ();
        var box = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);

        paned.pack1 (box, true, false);
        paned.pack2 (user_list, false, false);

        text_entry = new Gtk.Entry ();
        text_entry.hexpand = true;

        left_scrolled = new Utils.AutoScrolled ();
        left_scrolled.hscrollbar_policy = Gtk.PolicyType.NEVER;

        chat = new ChatDisplay ();

        topic_bar = new Widgets.Topic ();

        pseudo_label = new Gtk.Label ("");
        pseudo_label.set_alignment (1, (float) 0.5);
        pseudo_label.width_request = 80;

        var entry_eventbox = new Gtk.EventBox ();
        entry_eventbox.get_style_context ().add_class (Granite.StyleClass.CONTENT_VIEW);
        var entry_grid = new Gtk.Grid ();
        entry_grid.margin = 12;
        entry_grid.column_spacing = 12;
        entry_eventbox.add (entry_grid);

        entry_grid.attach (pseudo_label, 0, 0, 1, 1);
        entry_grid.attach (text_entry, 1, 0, 1, 1);
        left_scrolled.add_with_viewport (chat);

        box.pack_start (topic_bar, false);
        box.pack_start (left_scrolled);
        box.pack_start (entry_eventbox, false);

        loading_screen = new Granite.Widgets.EmbeddedAlert ();
        loading_screen.primary_text = _("Joining channel %s").printf (channel);
        loading_screen.working = true;

        action_join = new Gtk.Action ("join", "Join Channel", "Join this channel", null);

        not_joined_alert = new Granite.Widgets.EmbeddedAlert ();
        not_joined_alert.primary_text = _("Channel is not joined");
        not_joined_alert.secondary_text = _("Join the channel to receive and send messages.");
        not_joined_alert.message_type = Gtk.MessageType.WARNING;
        not_joined_alert.actions = {action_join};

        no_such_alert = new Granite.Widgets.EmbeddedAlert ();
        no_such_alert.primary_text = _("Cannot join channel");
        no_such_alert.secondary_text = _("An error has occurred, this channel does not exist.");
        no_such_alert.message_type = Gtk.MessageType.WARNING;
        no_such_alert.show_icon=  true;

        state = State.CHAT;

        this.show_all ();
    }

    void connect_callbacks () {
        text_entry.key_press_event.connect ((ev) => {
            if (ev.keyval == Gdk.Key.Tab) {
                var text = text_entry.text;

                if (text == null || text == "")
                    return false;

                var words = text.split (" ");
                var word =  words[words.length-1];

                if (word == null || word == "")
                    return false;

                string[] completion = null;
                foreach (var nick in user_list.nicks)
                    if (word.down () == nick[0:word.length].down ())
                        completion += nick;

                if (completion == null)
                    return true;

                if (completion.length > 1) {
                    chat.message_suggestions (completion);
                    return true;
                } else {
                    text_entry.buffer.delete_text (text_entry.cursor_position - word.length, word.length);

                    if (text_entry.text != "") {
                        text_entry.insert_at_cursor (completion[0] + " ");
                    } else {
                        text_entry.insert_at_cursor (completion[0] + ": ");
                    }
                    return true;
                }
            } else if (ev.keyval == Gdk.Key.Up) {
                if (history.has_older ())
                    text_entry.text = history.get_older ();
                return true;
            } else if (ev.keyval == Gdk.Key.Down) {
                if (history.has_newer ())
                    text_entry.text = history.get_newer ();
                return true;
            }

            return false;
        });

        text_entry.activate.connect (() => {
            if (text_entry.text != "") {
                history.append (text_entry.text);
                send_message (text_entry.text);
            }
        });

        action_join.activate.connect (() => {
            rejoin (); // signal
        });
    }

    public void list_users (string[] nicks, string[] prefixes)
        requires (nicks.length == prefixes.length) {

        for (int i = 0; i < nicks.length; i++) {
            switch (prefixes[i]) {
                case "@":
                    user_list.add_user (nicks[i], UserType.OPERATOR);
                    break;
                case "+":
                    user_list.add_user (nicks[i], UserType.VOICED);
                    break;
                default:
                    user_list.add_user (nicks[i], UserType.REGULAR);
                    break;
            }
        }
    }

    public bool has_user (string name) {
        return user_list.has_user (name);
    }

    public void rename_user (string? old_name, string new_name,
                             UserType type, bool silent = false) {

        if (old_name == null) {
            add_user (new_name, type, true);
            return;
        }

        remove_user (old_name, "", true);
        add_user (new_name, type, true);
        
        if (!silent)
            chat.message_nick (old_name, new_name);
    }

    public void add_user (string name, UserType type, bool silent = false) {
        user_list.add_user (name, type);
        if (!silent)
            chat.message_join (name);
    }

    public void remove_user (string name, string message, bool silent = false) {
        user_list.remove_user (name);
        if (!silent)
            chat.message_part (name, message);
    }
}
