/***
    Copyright (C) 2013 Cable Developers

    This program or library is free software; you can redistribute it
    and/or modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with this library; if not, write to the
    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301 USA.
***/

namespace Cable {
    public Gtk.ActionGroup main_actions;
}

public class Cable.Widgets.Window : Gtk.Window, View.Window {

    Granite.Widgets.SourceList server_list;
    Gtk.EventBox display_box;
    Gtk.Grid layout;
    Gtk.Grid server_grid;
    Granite.Widgets.ThinPaned paned;
    Widgets.Welcome welcome;

    Gtk.Toolbar channel_control;
    Gtk.ToolButton join_button;
    Gtk.ToolButton leave_button;
    Gtk.ToolButton prefs_button;

    Gtk.UIManager ui;

    private bool window_maximized = false;
    private int window_width = 0;
    private int window_height = 0;

    const string ui_string = """
        <ui>
            <popup name="MenuItemTool">
                <menuitem name="Join Channel" action="JoinChannel"/>
                <menuitem name="Leave Channel" action="LeaveChannel"/>
                <menuitem name="Edit Identity" action="EditCurrentId"/>
            </popup>
        </ui>
    """;

    public const string TITLE = "Cable";

    /**
     * Get or set the widget displayed next to the source list.
     */
    Gtk.Widget display_widget {
        get { return display_box.get_children ().nth_data (0); }
        set {
            var content = display_box.get_children ().nth_data (0);

            if (content != null)
                display_box.remove (content);

            display_box.add (value);
            display_box.show_all ();
        }
    }

    /**
     * Wether the welcome screen is shown or not.
     */
    public bool welcome_shown {
        get { return layout.get_child_at (0, 1) == welcome; }
        set {
            if (value) {
                var widget = layout.get_child_at (0, 1);
                if (widget != null)
                    layout.remove (widget);
                layout.attach (welcome, 0, 1, 1, 1);
                title = TITLE;
                welcome.focus_channel_entry ();
                this.show_all ();
            } else {
                var widget = layout.get_child_at (0, 1);
                if (widget != null)
                    layout.remove (widget);
                layout.attach (paned, 0, 1, 1, 1);
                this.show_all ();
            }
        }
    }

    public Gee.ArrayList <View.Server> servers {
        owned get {
            var _servers = new Gee.ArrayList <View.Server> ();
            foreach (var child in server_list.root.children)
                _servers.add (child as View.Server);
            return _servers;
        }
    }

        /**
     * Get the server of the currently selected channel.
     */
    public View.Server? current_server {
        get { return server_list.selected.parent as View.Server; }
    }

    /**
     * Get the currently selected channel.
     */
    public View.Chat? current_chat {
        get { return server_list.selected as View.Chat; }
        set { server_list.selected = value as Widgets.Chat; }
    }

    //public signal void edit_identity ();

    public Window (Cable.App app) {
        GLib.Object (application: app);
        init_actions ();
        init_layout ();

        server_list.item_selected.connect ((item) => {
            if (item != null) {
                display_widget = (current_chat as Widgets.Chat).room;
                selected (current_chat);
            }
        });

        welcome.done.connect ((a,n,c) => {
            join (a,n,c);
        });

        destroy.connect (() => {
            close ();
        });
    }

   void init_actions () {
        main_actions = new Gtk.ActionGroup ("MainActionGroup");
        main_actions.set_translation_domain (Constants.GETTEXT_PACKAGE);
        main_actions.add_actions (main_entries, this);

        ui = new Gtk.UIManager ();

        try {
            ui.add_ui_from_string (ui_string, -1);
        } catch (Error e) {
            error ("Couldn't load the UI: %s", e.message);
        }

        Gtk.AccelGroup accel_group = ui.get_accel_group ();
        add_accel_group (accel_group);

        ui.insert_action_group (main_actions, 0);
        ui.ensure_update ();
    }

    void init_layout () {
        this.title = TITLE;
        this.icon_name = "internet-chat";
        this.set_default_size (Settings.saved_state.window_width, Settings.saved_state.window_height);
        this.window_position = Gtk.WindowPosition.CENTER;

        switch (Settings.saved_state.window_state) {
            case Settings.SavedState.WindowState.MAXIMIZED:
                window_maximized = true;
                this.maximize ();
                break;
        }

        layout = new Gtk.Grid ();

        server_list = new Granite.Widgets.SourceList ();
        server_list.width_request = 150;

        display_box = new Gtk.EventBox ();
        paned = new Granite.Widgets.ThinPaned ();

        channel_control = new Gtk.Toolbar ();
        channel_control.set_icon_size (Gtk.IconSize.SMALL_TOOLBAR);

        join_button = main_actions.get_action ("JoinChannel").create_tool_item() as Gtk.ToolButton;
        leave_button = main_actions.get_action ("LeaveChannel").create_tool_item() as Gtk.ToolButton;
        prefs_button = main_actions.get_action ("Preferences").create_tool_item() as Gtk.ToolButton;

        var expander = new Gtk.SeparatorToolItem ();
        expander.set_draw (true);
        expander.set_expand (true);

        welcome = new Widgets.Welcome ();
        welcome.hexpand = true;
        welcome.vexpand = true;

        channel_control.get_style_context ().add_class (Gtk.STYLE_CLASS_INLINE_TOOLBAR);
        channel_control.add (join_button);
        channel_control.add (leave_button);
        channel_control.add (expander);
        channel_control.add (prefs_button);

        server_grid = new Gtk.Grid ();
        server_grid.attach (server_list, 0, 0, 1, 1);
        server_grid.attach (channel_control, 0, 1, 1, 1);

        paned.position = Settings.saved_state.column_server_width;
        paned.pack1 (server_grid, false, false);
        paned.pack2 (display_box, true, false);

        this.add (layout);
        this.show_all ();
    }

    private void close () {
        Settings.saved_state.window_width = window_width;
        Settings.saved_state.window_height = window_height;
        Settings.saved_state.column_server_width = paned.position;

        if (window_maximized)
            Settings.saved_state.window_state = Settings.SavedState.WindowState.MAXIMIZED;
        else
            Settings.saved_state.window_state = Settings.SavedState.WindowState.NORMAL;
    }

    void action_join_channel () {
        add_channel ();
    }

    void action_leave_channel () {
        remove_channel ();
    }

    void action_edit_identity () {
        /*edit_identity ();
        var dialog = new Dialogs.ManageIdentity (server_list.current_server.backend.identity);
        dialog.show_all ();*/
    }

    void action_preferences () {
        show_preferences ();
    }


    /**
     * Removes a server from the List.
     */
    public void remove_server (string server_id)
        /*requires (has_server (server_id))*/
        ensures (!has_server (server_id)) {

        var server = get_server_by_id (server_id);
        return_if_fail (server is Widgets.Server);
        server_list.root.remove (server as Widgets.Server);
    }

    /**
     * Add server.
     */
    public View.Server? add_server (string server_name, string server_id)
        requires (!has_server (server_id))
        ensures (has_server (server_id)) {

        var server = new Widgets.Server (server_name, server_id);
        server.expanded = true;
        server_list.root.add (server);
        return (View.Server) server;
    }

    static const Gtk.ActionEntry[] main_entries = {
        { "JoinChannel", "list-add-symbolic", N_("Join Channel…"), "<Control>j",
          N_("Join Channel…"), action_join_channel },

        { "LeaveChannel", "list-remove-symbolic", N_("Leave current Channel"), "<Control>l",
          N_("Leave current Channel"), action_leave_channel },

        { "EditCurrentId", "avatar-default-symbolic", N_("Edit current Identity…"), "<Control>i",
          N_("Edit current Identity…"), action_edit_identity },

        { "Preferences", "document-properties-symbolic", N_("Preferences"), null,
          N_("Preferences"), action_preferences }
    };

    public override bool configure_event (Gdk.EventConfigure event) {
        window_maximized = (get_window ().get_state () == Gdk.WindowState.MAXIMIZED);

        get_size (out window_width, out window_height);

        return base.configure_event (event);
    }
}
